import contextlib
from datetime import datetime

from silx.io.dictdump import dicttonx

with contextlib.suppress(ModuleNotFoundError):
    from HardwareRepository import HardwareRepository as HWR


class CreateH5Master:
    """Class to create HDF5 master file for dectris filewriter"""

    def __init__(self):
        self.file_template = None
        self._metadata = {}
        self.detector = {}
        self.detector_specific = {}
        self.beam = {}
        self.sample = {}
        self.sample_goniometer = {}
        self.sample_transformations = {}
        self.geometry = {}
        self.data = {}
        self.entry = {}
        self._pixel_size = ()
        self.read_only = None

    def init(self):
        """Initialise the static data, from the detector config"""
        # self._pixel_size = HWR.beamline.detector.get_pixel_size()
        self._pixel_size = (0.075, 0.075)
        self._metadata["description"] = "EIGER2 Si 9M".encode()
        self._metadata["detector_number"] = "E-18-0133, ESRF ID23".encode()
        self._metadata["sensor_material"] = "Si".encode()
        self._metadata["sensor_thickness"] = 0.00045
        self._metadata["detector_readout_time"] = 1e-7
        self.read_only = True
        # self._metadata["width"] = HWR.beamline.detector.get_width()
        # self._metadata["height"] = HWR.beamline.detector.get_height()
        self._metadata["width"] = 3108
        self._metadata["height"] = 3262

    def dynamic_values(self, start, osc_range, exptime, number_of_images):
        """Update the data structure with the specific data collection values
        Args:
            start (float): Oscillation start
            osc_range (float): Range for 1 image
            exptime (float): Exposure time for 1 image
            number_of_images (int): Number of images.
        """
        # beam_x, beam_y = HWR.beamline.detector.get_beam_position()
        beam_x = 1712.77
        beam_y = 2034.59
        # distance = HWR.beamline.detector.distance.get_value()/ 1000.0
        distance = 0.14104
        # data_collection_date = datetime.now().astimezone().isoformat()
        # 2023-02-23T16:33:53.537684+01:00
        self.detector.update(
            {
                "beam_center_x": beam_x,
                "beam_center_y": beam_y,
                "count_time": exptime,
                "frame_time": exptime + self._metadata["detector_readout_time"],
                "distance": distance,
                "nimages": number_of_images,
            }
        )
        if not self.read_only:
            self.detector_specific.update(
                {
                    "photon_energy": HWR.beamline.energy.get_value() * 1000,
                    "data_collection_date": datetime.now().astimezone().isoformat(),
                    "nimages": number_of_images,
                }
            )
            # self.beam["incident_wavelength"] = HWR.beamline.energy.get_wavelength()
            self.beam["incident_wavelength"] = 0.873128

        self.geometry.update(
            {
                "translation": {
                    "distances": [0.126571944, 0.152385525, -distance],
                }
            },
        )

        omega = []
        omega_end = []
        for img in range(number_of_images):
            omega.append(start + img * osc_range)
            omega_end.append(start + (img + 1) * osc_range)
        osc_range_total = number_of_images * osc_range
        self.sample_goniometer.update(
            {
                "omega": omega,
                "omega_end": omega_end,
                "omega_range_average": osc_range,
                "omega_range_total": osc_range_total,
            }
        )

        self.sample_transformations.update(
            {
                "omega": omega,
                "omega_end": omega_end,
                "omega_range_average": osc_range,
                "omega_range_total": osc_range_total,
                "translation": [0] * number_of_images,
            }
        )

        nfiles = int(number_of_images / 100)
        n_images_left = number_of_images % 100
        if n_images_left:
            nfiles += 1
        fname = "mesh-M_collect_test_0_1_"
        for i in range(1, nfiles + 1):
            self.data[f">data_{i:06d}"] = f"{fname}data_{i:06d}.h5::/entry/data/data"
        self.data["@signal"] = f"data_{i:06d}"

    def static_values(self):
        """Prepare the data structure, to be done at the begining of
        the datam collection
        """
        self.detector_specific = {
            "@NX_class": "NXcollection",
            "frame_count_time@units": "s",
            "frame_period@units": "s",
            "photon_energy@units": "eV",
            ## "compression": self.get_channel_object("compression_type").get_value(),
            "compression": "bslz4",
            "trigget_mode": "exts",
            "x_pixels_in_detector": self._metadata["width"],
            "y_pixels_in_detector": self._metadata["height"],
        }

        if self.read_only:
            self.detector_specific.update({"photon_energy": 14200})

        self.geometry = {
            "@NX_class": "NXgeometry",
            "orientation": {
                "@NX_class": "NXorientation",
                "value": [-1, 0, 0, 0, -1, 0],
            },
            "translation": {
                "@NX_class": "NXtranslation",
                "distances@units": "m",
            },
        }

        self.beam = {
            "@NX_class": "NXbeam",
            "incident_wavelength@units": "angstrom",
        }

        if self.read_only:
            ## self.detector_specific["photon_energy"] = HWR.beamline.energy.get_value()
            ## self.beam["incident_wavelength"] = HWR.beamline.energy.get_wavelength()
            self.detector_specific["photon_energy"] = 14200
            self.beam["incident_wavelength"] = 0.8731281579802835

        self.module = {
            "@NX_class": "NXdetector_module",
            "data_size": [self._metadata["width"], self._metadata["height"]],
            "fast_pixel_direction@depends_on": "/entry/instrument/detector/transformations/translation",  # noqa: E501
            "fast_pixel_direction@units": "m",
            "fast_pixel_direction": self._pixel_size[0] / 10000.0,
            "slow_pixel_direction@depends_on": "/entry/instrument/detector/transformations/translation",  # noqa: E501
            "slow_pixel_direction@units": "m",
            "slow_pixel_direction": self._pixel_size[0] / 10000.0,
        }
        self.detector = {
            "@NX_class": "NXdetector",
            "beam_center_x@units": "pixel",
            "beam_center_y@units": "pixel",
            "count_time@units": "s",
            "detector_readout_time@units": "s",
            "distance@units": "m",
            "frame_time@units": "s",
            "sensor_thickness@units": "m",
            "threshold_energy@units": "eV",
            "x_pixel_size": self._pixel_size[0] / 1000.0,
            "x_pixel_size@units": "m",
            "y_pixel_size": self._pixel_size[1] / 1000.0,
            "y_pixel_size@units": "m",
            "description": self._metadata["description"],
            "detector_number": self._metadata["detector_number"],
            "detector_readout_time": self._metadata["detector_readout_time"],
            "bit_depth_image": 32,
            "bit_depth_readout": 16,
            "countrate_correction_applied": 1,
            "sensor_material": self._metadata["sensor_material"],
            "sensor_thickness": self._metadata["sensor_thickness"],
            "threshold_energy": self.detector_specific["photon_energy"],
            "type": "HPC",
            "virtual_pixel_correction_applied": 1,
            "detectorSpecific": self.detector_specific,
            "geometry": self.geometry,
            "goniometer": {"@NX_class": "NXtransformations"},
            "module": self.module,
        }

        self.sample_goniometer = {
            "@NX_class": "NXtransformations",
            "omega@units": "degree",
            "omega_end@units": "degree",
            "omega_range_average@units": "degree",
            "omega_range_total@units": "degree",
        }

        self.sample_transformations = {
            "@NX_class": "NXtransformations",
            "omega@depends_on": "/entry/sample/transformations/translation",
            "omega@transformation_type": "rotation",
            "omega@units": "degree",
            "omega@offset": [0.0, 0.0, 0.0],
            "omega@vector": [-1.0, 0.0, 0.0],
            "omega_end@units": "degree",
            "omega_range_average@units": "degree",
            "omega_range_total@units": "degree",
            "translation@transformation_type": "translation",
            "translation@units": "m",
            "translation@depends_on": ".",
            "translation@offset": [0.0, 0.0, 0.0],
            "translation@vector": [0.0, 0.0, 0.0],
        }

        self.sample = {
            "@NX_class": "NXsample",
            "depends_on": "/entry/sample/transformations/omega",
            "beam": self.beam,
            "goniometer": self.sample_goniometer,
            "transformations": self.sample_transformations,
        }

        self.data = {"@NX_class": "NXdata"}

        self.entry = {
            "@default": "entry",
            "entry": {
                "@NX_class": "NXentry",
                "@default": "data",
                "data": self.data,
                "definition": "NXmx".encode(),
                "definition@version": "1.4",
                "instrument": {
                    "@NX_class": "NXinstrument",
                    "beam": self.beam,
                    "detector": self.detector,
                },
                "sample": self.sample,
            },
        }

    def write_master_file(self, fdir=None, fname=None):
        """Write the master file.
        Args:
            fdir (str): Directory to write te file.
            fname (str): file name.
        """
        fdir = fdir or "/home/esrf/beteva/hdf5_dectris.git/data"
        fname = fname or "ref-insu-insu_2_1_"
        nfiles = 1

        for i in range(1, nfiles + 1):
            self.data[f">data_{i:06d}"] = f"{fname}data_{i:06d}.h5::/entry/data/data"
        self.data["@signal"] = f"data_{i:06d}"
        dicttonx(self.entry, f"{fdir}/{fname}master_new.h5", mode="w")



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import copy
import logging
import os
import time

from XSDataCommon import (
    XSDataAngle,
    XSDataBoolean,
    XSDataDouble,
    XSDataFlux,
    XSDataImage,
    XSDataInteger,
    XSDataLength,
    XSDataSize,
    XSDataString,
    XSDataTime,
    XSDataWavelength,
)
from XSDataMXCuBEv1_4 import (
    XSDataInputMXCuBE,
    XSDataMXCuBEDataSet,
    XSDataResultMXCuBE,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.EDNACharacterisation import EDNACharacterisation
from mxcubecore.model import queue_model_enumerables as qme
from mxcubecore.model import queue_model_objects as qmo

# from edna_test_data import EDNA_DEFAULT_INPUT
# from edna_test_data import EDNA_TEST_DATA


class P11EDNACharacterisation(EDNACharacterisation):
    def __init__(self, name):
        super().__init__(name)

        self.collect_obj = None
        self.result = None
        self.edna_default_file = None
        self.start_edna_command = None

    def _run_edna(self, input_file, results_file, process_directory):
        # First submit MOSFLM job
        self.mosflm_maxwell()

        """Starts EDNA"""
        msg = "Starting EDNA characterisation using xml file %s" % input_file
        logging.getLogger("queue_exec").info(msg)

        args = (self.start_edna_command, input_file, results_file, process_directory)
        # subprocess.call("%s %s %s %s" % args, shell=True)

        # Test run DESY
        self.edna_maxwell(process_directory, input_file, results_file)

        self.result = None
        if os.path.exists(results_file):
            self.result = XSDataResultMXCuBE.parseFile(results_file)

        return self.result

    def mosflm_maxwell(self):
        self.log.debug("==== MOSFLM CHARACTERISATION IN PROGRESS ==========")

        # Retrieve resolution value and number of frames
        resolution = HWR.beamline.resolution.get_value()
        frames = HWR.beamline.collect.current_dc_parameters["oscillation_sequence"][0][
            "number_of_images"
        ]

        # Fetch the latest local master image filename
        latest_h5_filename = HWR.beamline.detector.get_latest_local_master_image_name()

        # Prepare local and remote image directory paths
        image_dir_local, filename = os.path.split(latest_h5_filename)
        remote_base_path = HWR.beamline.session.get_beamtime_metadata()[2]
        image_dir = image_dir_local.replace("/gpfs/current", remote_base_path)
        process_dir_local = image_dir_local.replace("/raw/", "/processed/")
        process_dir = image_dir.replace("/raw/", "/processed/")
        mosflm_path_local = os.path.join(process_dir_local, "mosflm")
        mosflm_path = os.path.join(process_dir, "mosflm")

        self.log.debug(f'============MOSFLM======== mosflm_path="{mosflm_path}"')
        self.log.debug(
            f'============MOSFLM======== mosflm_path_local="{mosflm_path_local}"'
        )

        # Create local MOSFLM directory
        try:
            self.mkdir_with_mode(mosflm_path_local, mode=0o777)
            self.log.debug("=========== MOSFLM ============ Mosflm directory created")
        except OSError as e:
            self.log.debug(f"Cannot create mosflm directory: {e}")

        # Update datasets.txt file for presenterd
        base_process_dir = HWR.beamline.collect.base_dir(process_dir_local, "processed")
        datasets_file = os.path.join(base_process_dir, "datasets.txt")
        try:
            with open(datasets_file, "a") as file:
                file.write(
                    f"{mosflm_path_local.split('/gpfs/current/processed/')[1]}\n"
                )
        except (OSError, RuntimeWarning) as err:
            self.log.debug(f"Cannot write to datasets.txt: {err}")

        # Create SBATCH command
        ssh = HWR.beamline.session.get_ssh_command()
        log_file_path = (
            mosflm_path.replace(remote_base_path, "/beamline/p11/current")
            + "/mosflm.log"
        )
        sbatch = HWR.beamline.session.get_sbatch_command(
            jobname_prefix="mosflm", logfile_path=log_file_path
        )

        # Formulate processing command
        cmd = (
            f"/asap3/petra3/gpfs/common/p11/processing/mosflm_sbatch.sh "
            f"{image_dir} {filename} {mosflm_path.replace(remote_base_path, '/beamline/p11/current')} "
            f"{frames} {resolution}"
        )

        # Log and execute the command
        self.log.debug(f'=======MOSFLM========== ssh="{ssh}"')
        self.log.debug(f'=======MOSFLM========== sbatch="{sbatch}"')
        self.log.debug(f'=======MOSFLM========== executing process cmd="{cmd}"')
        full_cmd = f'{ssh} "{sbatch} --wrap \\"{cmd}\\""'
        self.log.debug(f"=======MOSFLM========== {full_cmd}")

        os.system(full_cmd)

    def edna_maxwell(self, process_directory, inputxml, outputxml):
        """
        Executes a command on a remote cluster using SSH and SBATCH for EDNA processing.

        :param process_directory: Directory where the processing will take place.
        :param inputxml: Path to the input XML file.
        :param outputxml: Path to the output XML file that will be generated.
        """
        self.log.debug(f'=======EDNA========== PROCESS DIRECTORY="{process_directory}"')
        self.log.debug(f'=======EDNA========== IN (XML INPUT)="{inputxml}"')
        self.log.debug(f'=======EDNA========== OUT (XML OUTPUT)="{outputxml}"')

        ssh = HWR.beamline.session.get_ssh_command()
        sbatch = HWR.beamline.session.get_sbatch_command(
            jobname_prefix="edna",
            logfile_path=process_directory.replace("/gpfs", "/beamline/p11")
            + "/edna.log",
        )

        inxml = inputxml.replace("/gpfs", "/beamline/p11")
        outxml = outputxml.replace(
            "/gpfs/current/processed", "/beamline/p11/current/processed"
        )

        processpath = process_directory.replace(
            "/gpfs/current/processed", "/beamline/p11/current/processed"
        )

        self.log.debug(
            f'=======EDNA========== CLUSTER PROCESS DIRECTORY="{processpath}"'
        )
        self.log.debug(f'=======EDNA========== CLUSTER IN="{inxml}"')
        self.log.debug(f'=======EDNA========== CLUSTER OUT="{outxml}"')
        self.log.debug(f'=======EDNA========== ssh="{ssh}"')
        self.log.debug(f'=======EDNA========== sbatch="{sbatch}"')

        cmd = f"/asap3/petra3/gpfs/common/p11/processing/edna_sbatch.sh {inxml} {outxml} {processpath}"

        self.log.debug(f'=======EDNA========== executing process cmd="{cmd}"')
        self.log.debug(f'=======EDNA========== {ssh} "{sbatch} --wrap \\"{cmd}\\""')
        logging.info(f'{ssh} "{sbatch} --wrap \\"{cmd}\\""')

        waitforxml = outputxml.replace("/raw/", "/processed/")
        self.log.debug(f"=======EDNA========== WAITING FOR OUTPUTXML IN {waitforxml}")

        self.log.debug(f'=======EDNA Script========== {ssh} "{cmd}"')
        os.system(f'{ssh} "{cmd}"')

        self.wait_for_file(waitforxml, timeout=60)

        # Update datasets.txt file for presenterd
        base_process_dir = "/gpfs/current/processed"
        datasets_file = os.path.join(base_process_dir, "datasets.txt")
        try:
            with open(datasets_file, "a") as file:
                file.write(
                    f"{process_directory.split('/gpfs/current/processed/')[1]}\n"
                )
        except (OSError, RuntimeWarning) as err:
            self.log.debug(f"Cannot write to datasets.txt: {err}")

    def wait_for_file(self, file_path, timeout=60, check_interval=1):
        start_time = time.time()
        while not os.path.exists(file_path):
            elapsed_time = time.time() - start_time
            if elapsed_time >= timeout:
                self.log.debug(
                    f"Timeout reached. File '{file_path}' not found within {timeout} seconds."
                )
                raise RuntimeWarning(
                    f"Timeout reached. File '{file_path}' not found within {timeout} seconds."
                )
            self.log.info(
                f"Waiting for file '{file_path}'... ({elapsed_time:.1f}/{timeout} seconds elapsed)"
            )
            time.sleep(check_interval)
        self.log.info(f"File '{file_path}' found.")

    def input_from_params(self, data_collection, char_params):
        edna_input = XSDataInputMXCuBE.parseString(self.edna_default_input)

        if data_collection.id:
            edna_input.setDataCollectionId(XSDataInteger(data_collection.id))

        # Beam object
        beam = edna_input.getExperimentalCondition().getBeam()

        try:
            transmission = HWR.beamline.transmission.get_value()
            beam.setTransmission(XSDataDouble(transmission))
        except AttributeError:
            self.log.debug("EDNACharacterisation. transmission not saved ")
            self.log.exception("")

        try:
            wavelength = HWR.beamline.energy.get_wavelength()
            beam.setWavelength(XSDataWavelength(wavelength))
        except AttributeError:
            self.log.exception("")

        try:
            # Flux get_value() subsequently executing measure_flux() to fill in the dictionary

            beam.setFlux(XSDataFlux(HWR.beamline.flux.get_value()))
        except AttributeError:
            self.log.exception("")

        try:
            min_exp_time = self.collect_obj.detector_hwobj.get_exposure_time_limits()[0]
            beam.setMinExposureTimePerImage(XSDataTime(min_exp_time))
        except AttributeError:
            self.log.exception("")

        try:
            beamsize = self.collect_obj.beam_info_hwobj.get_beam_size()

            if None not in beamsize:
                beam.setSize(
                    XSDataSize(
                        x=XSDataLength(float(beamsize[0])),
                        y=XSDataLength(float(beamsize[1])),
                    )
                )
        except AttributeError:
            self.log.exception("")

        # Optimization parameters
        diff_plan = edna_input.getDiffractionPlan()

        aimed_i_sigma = XSDataDouble(char_params.aimed_i_sigma)
        aimed_completness = XSDataDouble(char_params.aimed_completness)
        aimed_multiplicity = XSDataDouble(char_params.aimed_multiplicity)
        aimed_resolution = XSDataDouble(char_params.aimed_resolution)

        complexity = char_params.strategy_complexity
        complexity = XSDataString(qme.STRATEGY_COMPLEXITY[complexity])

        permitted_phi_start = XSDataAngle(char_params.permitted_phi_start)
        _range = char_params.permitted_phi_end - char_params.permitted_phi_start
        rotation_range = XSDataAngle(_range)

        if char_params.aimed_i_sigma:
            diff_plan.setAimedIOverSigmaAtHighestResolution(aimed_i_sigma)

        if char_params.aimed_completness:
            diff_plan.setAimedCompleteness(aimed_completness)

        if char_params.use_aimed_multiplicity:
            diff_plan.setAimedMultiplicity(aimed_multiplicity)

        if char_params.use_aimed_resolution:
            diff_plan.setAimedResolution(aimed_resolution)

        diff_plan.setComplexity(complexity)
        diff_plan.setStrategyType(XSDataString(char_params.strategy_program))

        if char_params.use_permitted_rotation:
            diff_plan.setUserDefinedRotationStart(permitted_phi_start)
            diff_plan.setUserDefinedRotationRange(rotation_range)

        # Vertical crystal dimension
        sample = edna_input.getSample()
        sample.getSize().setY(XSDataLength(char_params.max_crystal_vdim))
        sample.getSize().setZ(XSDataLength(char_params.min_crystal_vdim))

        # Radiation damage model
        sample.setSusceptibility(XSDataDouble(char_params.rad_suscept))
        sample.setChemicalComposition(None)
        sample.setRadiationDamageModelBeta(XSDataDouble(char_params.beta / 1e6))
        sample.setRadiationDamageModelGamma(XSDataDouble(char_params.gamma / 1e6))

        diff_plan.setForcedSpaceGroup(XSDataString(char_params.space_group))

        # Characterisation type - Routine DC
        if char_params.use_min_dose:
            pass

        if char_params.use_min_time:
            time = XSDataTime(char_params.min_time)
            diff_plan.setMaxExposureTimePerDataCollection(time)

        # Account for radiation damage
        if char_params.induce_burn:
            self._modify_strategy_option(diff_plan, "-DamPar")

        # Characterisation type - SAD
        if char_params.opt_sad:
            if char_params.auto_res:
                diff_plan.setAnomalousData(XSDataBoolean(True))
            else:
                diff_plan.setAnomalousData(XSDataBoolean(False))
                self._modify_strategy_option(diff_plan, "-SAD yes")
                diff_plan.setAimedResolution(XSDataDouble(char_params.sad_res))
        else:
            diff_plan.setAnomalousData(XSDataBoolean(False))

        # Data set
        data_set = XSDataMXCuBEDataSet()
        acquisition_parameters = data_collection.acquisitions[0].acquisition_parameters
        path_template = data_collection.acquisitions[0].path_template

        # NB!: path_template content is in queue_model_objects.py
        logging.info(
            "======= Characterisation path template directory ====%s",
            path_template.directory,
        )
        logging.info(
            "======= Characterisation path template process_directory ====%s",
            path_template.process_directory,
        )
        logging.info(
            "======= Characterisation path template xds_dir ====%s",
            path_template.xds_dir,
        )
        logging.info(
            "======= Characterisation path template base_prefix ====%s",
            path_template.base_prefix,
        )
        logging.info(
            "======= Characterisation path template mad_prefix ====%s",
            path_template.mad_prefix,
        )
        logging.info(
            "======= Characterisation path template reference_image_prefix ====%s",
            path_template.reference_image_prefix,
        )
        logging.info(
            "======= Characterisation path template wedge_prefix ====%s",
            path_template.wedge_prefix,
        )
        logging.info(
            "======= Characterisation path template run_number ====%s",
            str(path_template.run_number),
        )
        logging.info(
            "======= Characterisation path template suffix ====%s", path_template.suffix
        )
        logging.info(
            "======= Characterisation path template precision ====%s",
            str(path_template.precision),
        )
        logging.info(
            "======= Characterisation path template start_num ====%s",
            str(path_template.start_num),
        )
        logging.info(
            "======= Characterisation path template num_files ====%s",
            str(path_template.num_files),
        )
        logging.info(
            "======= Characterisation path template compression ====%s",
            path_template.compression,
        )

        # Make sure there is a proper path conversion between different mount points

        # This is where the folder with h5 files is located.
        # It is straightforward to construct it from xds_dir
        image_dir = path_template.xds_dir.replace(
            "/gpfs/current/processed/", "/beamline/p11/current/raw/"
        ).replace("/edna", "")

        # This is needed because EDNA needs to concert eiger2cbf in place
        # The directory is created by local user, EDNA process is executed by another without
        # access rights.
        raw_char_directory = path_template.xds_dir.replace(
            "/gpfs/current/processed/", "/gpfs/current/raw/"
        ).replace("/edna", "")
        os.system(f"chmod +777 {raw_char_directory}")

        # Here is actual path to the *.h5 data are defined:
        # It is expected to be %5d (4 zeros), but files are generated %6d (5 leading zeros).
        # It also requires _data_%6d.h5
        # TODO: Fix elsewhere.
        path_template.precision = 6

        path_str = os.path.join(image_dir, path_template.get_image_file_name())

        logging.info(
            "======= Characterisation path of what image filename is expected ====%s",
            path_template.get_image_file_name(),
        )
        logging.info(
            "======= Characterisation path where to search for images ====%s", path_str
        )

        # NB!: Directories at this point are created elsewhere (data_collection_hook)
        # xds_dir at this point already has all the needed substitutions.
        characterisation_dir = path_template.xds_dir

        # pattern = r"(_\d{6}\.h5)$"

        for img_num in range(int(acquisition_parameters.num_images)):
            image_file = XSDataImage()
            path = XSDataString()
            path.value = path_str % (img_num + 1)
            image_file.path = path
            (image_file.path.value,)
            image_file.number = XSDataInteger(img_num + 1)
            data_set.addImageFile(image_file)

        edna_input.addDataSet(data_set)
        edna_input.process_directory = characterisation_dir

        return edna_input

    def dc_from_output(self, edna_result, reference_image_collection):
        data_collections = []

        crystal = copy.deepcopy(reference_image_collection.crystal)
        ref_proc_params = reference_image_collection.processing_parameters
        processing_parameters = copy.deepcopy(ref_proc_params)

        try:
            char_results = edna_result.getCharacterisationResult()
            edna_strategy = char_results.getStrategyResult()
            collection_plan = edna_strategy.getCollectionPlan()[0]
            wedges = collection_plan.getCollectionStrategy().getSubWedge()
        except Exception:
            self.log.exception("")
        else:
            try:
                resolution = (
                    collection_plan.getStrategySummary().getResolution().getValue()
                )
                resolution = round(resolution, 3)
            except AttributeError:
                resolution = None

            try:
                transmission = (
                    collection_plan.getStrategySummary().getAttenuation().getValue()
                )
                transmission = round(transmission, 2)
            except AttributeError:
                transmission = None

            try:
                screening_id = edna_result.getScreeningId().getValue()
            except AttributeError:
                screening_id = None

            # GB merging wedges wedges
            osc_very_end = (
                wedges[-1]
                .getExperimentalCondition()
                .getGoniostat()
                .getRotationAxisEnd()
                .getValue()
            )

            for i in range(0, 1):  # GB len(wedges)):
                wedge = wedges[i]
                exp_condition = wedge.getExperimentalCondition()
                goniostat = exp_condition.getGoniostat()
                beam = exp_condition.getBeam()

                acq = qmo.Acquisition()
                acq.acquisition_parameters = (
                    HWR.beamline.get_default_acquisition_parameters()
                )
                acquisition_parameters = acq.acquisition_parameters

                acquisition_parameters.centred_position = (
                    reference_image_collection.acquisitions[
                        0
                    ].acquisition_parameters.centred_position
                )

                acq.path_template = HWR.beamline.get_default_path_template()

                # Use the same path template as the reference_collection
                # and update the members the needs to be changed. Keeping
                # the directories of the reference collection.
                ref_pt = reference_image_collection.acquisitions[0].path_template

                acq.path_template = copy.deepcopy(ref_pt)

                # This part was removing the "raw" from final directory and failing to collect data from EDNA-generated diffraction plan.
                # Keep it here until it is clear that it is not propagating further.
                # acq.path_template.directory = "/".join(
                #     ref_pt.directory.split("/")[0:-2]
                # )

                # acq.path_template.wedge_prefix = "w" + str(i + 1)
                acq.path_template.reference_image_prefix = str()

                if resolution:
                    acquisition_parameters.resolution = resolution

                if transmission:
                    acquisition_parameters.transmission = transmission

                if screening_id:
                    acquisition_parameters.screening_id = screening_id

                try:
                    acquisition_parameters.osc_start = (
                        goniostat.getRotationAxisStart().getValue()
                    )
                except AttributeError:
                    self.log.exception("")
                """ GB:
                try:
                    acquisition_parameters.osc_end = (
                        goniostat.getRotationAxisEnd().getValue()
                    )
                except AttributeError:
                    pass
                """
                acquisition_parameters.osc_end = osc_very_end

                try:
                    acquisition_parameters.osc_range = (
                        goniostat.getOscillationWidth().getValue()
                    )
                except AttributeError:
                    self.log.exception("")

                try:
                    num_images = int(
                        abs(
                            acquisition_parameters.osc_end
                            - acquisition_parameters.osc_start
                        )
                        / acquisition_parameters.osc_range
                    )

                    acquisition_parameters.first_image = 1
                    acquisition_parameters.num_images = num_images
                    acq.path_template.num_files = num_images
                    acq.path_template.start_num = 1

                except AttributeError:
                    self.log.exception("")

                try:
                    acquisition_parameters.transmission = (
                        beam.getTransmission().getValue()
                    )
                except AttributeError:
                    self.log.exception("")

                try:
                    acquisition_parameters.energy = round(
                        (123_984.0 / beam.getWavelength().getValue()) / 10000.0, 4
                    )
                except AttributeError:
                    self.log.exception("")

                try:
                    acquisition_parameters.exp_time = beam.getExposureTime().getValue()
                except AttributeError:
                    self.log.exception("")

                dc = qmo.DataCollection([acq], crystal, processing_parameters)
                data_collections.append(dc)

        return data_collections

    def mkdir_with_mode(self, directory, mode):
        """
        The function creates a directory with a specified mode if it does not already exist.

        :param directory: The "directory" parameter is the path of the directory that you want to
        create. It can be an absolute path or a relative path
        :param mode: The "mode" parameter in the above code refers to the permissions that will be set
        for the newly created directory. It is an optional parameter that specifies the access mode for
        the directory. The access mode is a numeric value that represents the permissions for the
        directory
        """
        if not os.path.isdir(directory):
            oldmask = os.umask(000)
            os.makedirs(directory, mode=mode)
            os.umask(oldmask)
            # self.checkPath(directory,force=True)

            self.log.debug("local directory created")

    def get_html_report(self, edna_result):
        """
        Args:
            output (EDNAResult) EDNAResult object

        Returns:
            (str) The path to the html result report generated by the characterisation
            software
        """
        html_report = None

        try:
            html_report = str(edna_result.getHtmlPage().getPath().getValue())
            html_report = html_report.replace("/beamline/p11", "/gpfs")

        except AttributeError:
            self.log.exception("")

        return html_report



import logging

from devtools import debug
from pydantic import (
    Field,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ESRF.queue_entry.ssx_base_queue_entry import (
    BaseUserCollectionParameters,
    SsxBaseQueueEntry,
    SsxBaseQueueTaskParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class SSXUserCollectionParameters(BaseUserCollectionParameters):
    motor_top_left_x: float = Field(7, gt=-100, lt=100)
    motor_top_left_y: float = Field(3, gt=-100, lt=100)
    motor_top_left_z: float = Field(0.08, gt=-100, lt=100)

    motor_top_right_x: float = Field(32, gt=-100, lt=100)
    motor_top_right_y: float = Field(3, gt=-100, lt=100)
    motor_top_right_z: float = Field(0.08, gt=-100, lt=100)

    motor_bottom_left_x: float = Field(7, gt=-100, lt=100)
    motor_bottom_left_y: float = Field(28, gt=-100, lt=100)
    motor_bottom_left_z: float = Field(0.08, gt=-100, lt=100)

    nb_samples_per_line: int = Field(500, gt=0, lt=1000)
    nb_lines: int = Field(3000, gt=0, lt=10000)

    class Config:
        extra: "ignore"


class SsxFoilColletionTaskParameters(SsxBaseQueueTaskParameters):
    user_collection_parameters: SSXUserCollectionParameters


class SsxFoilCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class SsxBigFoilCollectionQueueEntry(SsxBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = SsxFoilCollectionQueueModel
    DATA_MODEL = SsxFoilColletionTaskParameters
    NAME = "SSX Big Foil Collection"
    REQUIRES = ["point", "line", "no_shape", "chip", "grid"]

    def __init__(self, view, data_model: SsxFoilCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)

    def execute(self):
        super().execute()
        debug(self._data_model._task_data)
        params = self._data_model._task_data.user_collection_parameters

        exp_time = self._data_model._task_data.user_collection_parameters.exp_time
        fname_prefix = self._data_model._task_data.path_parameters.prefix
        num_images = params.nb_samples_per_line * params.nb_lines
        reject_empty_frames = (
            self._data_model._task_data.user_collection_parameters.reject_empty_frames
        )
        self._data_model._task_data.collection_parameters.num_images = num_images
        data_root_path = self.get_data_path()

        HWR.beamline.diffractometer.wait_ready()
        HWR.beamline.diffractometer.set_phase("DataCollection", wait=True, timeout=120)

        self.take_pedestal()

        logging.getLogger("user_level_log").info("Preparing detector")
        HWR.beamline.detector.prepare_acquisition(
            num_images,
            exp_time,
            data_root_path,
            fname_prefix,
            dense_skip_nohits=reject_empty_frames,
        )
        HWR.beamline.detector.wait_ready()

        fname_prefix = self._data_model._task_data.path_parameters.prefix
        fname_prefix += f"_foil_"

        region = (
            params.motor_top_left_x,
            params.motor_top_left_y,
            params.motor_top_left_z,
            params.motor_top_right_x,
            params.motor_top_right_y,
            params.motor_top_right_z,
            params.motor_bottom_left_x,
            params.motor_bottom_left_y,
            params.motor_bottom_left_z,
        )

        self.start_processing("FOIL")

        logging.getLogger("user_level_log").info(f"Defining region {region}")

        HWR.beamline.diffractometer.prepare_ssx_grid_scan(
            params.sub_sampling, *region, params.nb_samples_per_line, params.nb_lines
        )

        if HWR.beamline.control.safshut_oh2.state.name != "OPEN":
            logging.getLogger("user_level_log").info(f"Opening OH2 safety shutter")
            HWR.beamline.control.safshut_oh2.open()

        HWR.beamline.detector.start_acquisition()
        logging.getLogger("user_level_log").info(
            "Detector ready, waiting for trigger ..."
        )

        logging.getLogger("user_level_log").info(f"Acquiring {num_images}")

        try:
            HWR.beamline.diffractometer.start_ssx_scan()
            HWR.beamline.diffractometer.wait_ready()
        except:
            msg = "Diffractometer failed! Waiting for detector to finish"
            logging.getLogger("user_level_log").error(msg)
            HWR.beamline.detector.wait_ready()
            raise

        if HWR.beamline.control.safshut_oh2.state.name == "OPEN":
            HWR.beamline.control.safshut_oh2.close()
            logging.getLogger("user_level_log").info("shutter closed")

        HWR.beamline.detector.wait_ready()
        logging.getLogger("user_level_log").info(f"Acquired {region}")

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        super().stop()



import time
from typing import Optional

import gevent

from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor
from mxcubecore.HardwareObjects.LNLS.EPICS.EPICSActuator import EPICSActuator


class EPICSMotor(EPICSActuator, AbstractMotor):
    MOTOR_DMOV = "dmov"
    MOTOR_STOP = "stop"
    MOTOR_VELO = "velo"
    MOTOR_ACCL = "accl"
    MOTOR_HLM = "hlm"
    MOTOR_LLM = "llm"
    MOTOR_EGU = "egu"
    MOTOR_PREC = "prec"

    def _instantiate_attributes(self):
        pvname = self.get_channel_object("").command.pv_name
        self.add_channel({"type": "epics", "name": self.ACTUATOR_VAL}, pvname + ".VAL")
        self.add_channel(
            {"type": "epics", "polling": 200, "name": self.ACTUATOR_RBV},
            pvname + ".RBV",
        )
        self.add_channel({"type": "epics", "name": self.MOTOR_DMOV}, pvname + ".DMOV")
        self.add_channel({"type": "epics", "name": self.MOTOR_STOP}, pvname + ".STOP")
        self.add_channel({"type": "epics", "name": self.MOTOR_VELO}, pvname + ".VELO")
        self.add_channel({"type": "epics", "name": self.MOTOR_ACCL}, pvname + ".ACCL")
        self.add_channel({"type": "epics", "name": self.MOTOR_HLM}, pvname + ".HLM")
        self.add_channel({"type": "epics", "name": self.MOTOR_LLM}, pvname + ".LLM")
        self.add_channel({"type": "epics", "name": self.MOTOR_EGU}, pvname + ".EGU")
        self.add_channel({"type": "epics", "name": self.MOTOR_PREC}, pvname + ".PREC")

    def init(self):
        self._motor_channels = {}
        self._instantiate_attributes()
        self.get_limits()
        self.get_velocity()
        self.get_precision()
        super().init()

    def wait_ready(self, timeout: Optional[float] = None):
        self._ready_event.clear()
        timeout = abs(self.get_value() - self.setpoint) / self.get_velocity()
        timeout += 2 * self.get_acceleration()
        # Timeout tolerance
        timeout += 10
        is_set = self._ready_event.is_set()
        try:
            with gevent.Timeout(timeout, exception=TimeoutError):
                while not is_set:
                    is_set = self._ready_event.is_set()
                    if self.done_movement() and not self.hasnt_arrived(self.setpoint):
                        self._ready_event.set()
                    time.sleep(0.15)
        except TimeoutError:
            pvname = self.get_channel_object("").command.pv_name
            self.print_log(
                level="error",
                msg=f"{pvname} motion has timed out.",
            )
        self.update_state(self.STATES.READY)

    def abort(self):
        self.set_channel_value(self.MOTOR_STOP, 1)
        super().abort()

    def get_limits(self):
        try:
            low_limit = float(self.get_channel_value(self.MOTOR_LLM))
            high_limit = float(self.get_channel_value(self.MOTOR_HLM))
            self._nominal_limits = (low_limit, high_limit)
        except ValueError:
            self._nominal_limits = (None, None)
        if self._nominal_limits in [(0, 0), (float("-inf"), float("inf"))]:
            # Treat infinite limits
            self._nominal_limits = (None, None)
        return self._nominal_limits

    def get_velocity(self):
        self._velocity = self.get_channel_value(self.MOTOR_VELO)
        return self._velocity

    def get_precision(self):
        self._tolerance = 10 ** (-self.get_channel_value(self.MOTOR_PREC))

    def set_velocity(self, value):
        self.set_channel_value(self.MOTOR_VELO, value)
        self._velocity = value

    def get_acceleration(self):
        self._acceleration = self.get_channel_value(self.MOTOR_ACCL)
        return self._acceleration

    def done_movement(self):
        dmov = self.get_channel_value(self.MOTOR_DMOV)
        return bool(dmov)


class EPICSMotorDetachable(EPICSMotor):
    """
    Class that implements an interface of motor record IOC and accepts
    a presence PV, that indicated if the motor is or is not present on the
    facility. If it is present, it actuates as an EpicsMotor, if it is not,
    then the widget receives the state OFF and remains disabled.

    YAML Example
    ------------

    %YAML 1.2
    ---
    class: LNLS.EPICS.EPICSMotor.EPICSMotorDetachable
    epics:
    "MOTOR_PV":
        channels:
        '':
    'PRESENCE_PREFIX':
        channels:
        'presence':
            suffix: "PRESENCE_SUFFIX"
    configuration:
    is_absent_value: 0

    """

    MOTOR_PRESENCE_RBV = "presence"

    def init(self):
        super().init()
        self.get_value()

    def get_value(self):
        value = self.get_property("is_absent_value")
        current_value = int(self.get_channel_value(self.MOTOR_PRESENCE_RBV))
        if current_value == value:
            self.update_state(self.STATES.OFF)
        return super().get_value()



import logging

import gevent
from pydantic import (
    Field,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ESRF.queue_entry.ssx_base_queue_entry import (
    BaseUserCollectionParameters,
    SsxBaseQueueEntry,
    SsxBaseQueueTaskParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class LaserInjectorUserCollectionParameters(BaseUserCollectionParameters):
    num_images: int = Field(1000, gt=0, lt=10000000)
    take_pedestal: bool = Field(True)

    class Config:
        extra: "ignore"
        use_enum_values: True


class LaserInjectorColletionTaskParameters(SsxBaseQueueTaskParameters):
    user_collection_parameters: LaserInjectorUserCollectionParameters


class SsxLaserInjectorCollectionQueueModel(DataCollection):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SsxLaserInjectorCollectionQueueEntry(SsxBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = SsxLaserInjectorCollectionQueueModel
    DATA_MODEL = LaserInjectorColletionTaskParameters
    NAME = "SSX LaserInjector Collection"
    REQUIRES = ["point", "line", "no_shape", "chip", "grid"]

    def __init__(self, view, data_model: SsxLaserInjectorCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)
        self.__scanning = False

    def execute(self):
        super().execute()

        exp_time = self._data_model._task_data.user_collection_parameters.exp_time
        fname_prefix = self._data_model._task_data.path_parameters.prefix
        num_images = self._data_model._task_data.user_collection_parameters.num_images
        sub_sampling = (
            self._data_model._task_data.user_collection_parameters.sub_sampling
        )
        reject_empty_frames = (
            self._data_model._task_data.user_collection_parameters.reject_empty_frames
        )

        delay = HWR.beamline.diffractometer.get_ssx_delay()
        ssx_laser_scan_method = HWR.beamline.diffractometer.get_ssx_scan_method()

        data_root_path = self.get_data_path()

        HWR.beamline.diffractometer.set_phase("DataCollection")

        self.take_pedestal()

        HWR.beamline.detector.prepare_acquisition(
            num_images,
            exp_time,
            data_root_path,
            fname_prefix,
            dense_skip_nohits=reject_empty_frames,
        )

        self.start_processing("INJECTOR")

        if HWR.beamline.control.safshut_oh2.state.name != "OPEN":
            logging.getLogger("user_level_log").info(f"Opening OH2 safety shutter")
            HWR.beamline.control.safshut_oh2.open()

        HWR.beamline.diffractometer.wait_ready()
        HWR.beamline.detector.wait_ready()

        logging.getLogger("user_level_log").info(
            f"Laser scan method {ssx_laser_scan_method}"
        )
        logging.getLogger("user_level_log").info(f"Laser delay {delay}")
        logging.getLogger("user_level_log").info(f"Acquiring ...")
        HWR.beamline.detector.start_acquisition()

        try:
            HWR.beamline.diffractometer.start_still_ssx_scan(num_images)
        except:
            err_msg = "Diffractometer scan failed ..."
            logging.getLogger("user_level_log").exception(err_msg)
            HWR.beamline.detector.stop_acquisition()
            raise

        self.__scanning = True

        logging.getLogger("user_level_log").info("Waiting for scan to finish ...")

        try:
            HWR.beamline.diffractometer.wait_ready()
            logging.getLogger("user_level_log").info("Scan finished ...")
        finally:
            self.__scanning = False

            HWR.beamline.detector.wait_ready()
            acquired = HWR.beamline.detector.get_acquired_frames()
            logging.getLogger("user_level_log").info(f"Acquired {acquired} images")

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        if self.__scanning:
            logging.getLogger("user_level_log").info("Stopping diffractometer ...")
            HWR.beamline.diffractometer.abort_cmd()
            gevent.sleep(5)
            HWR.beamline.diffractometer.wait_ready()

        super().stop()



"""
TITLE
MultiplePositions Hardware Object

DESCRIPTION
This object manages the movement of several motors to predefined positions.

<username> : name of the multiplepositions object
<mode>     : there is two ways of managing the change of predefined positions
              absolute: change the value of a predefined position by another
                        absolute position
              relative: do not change the absolute value of the predefined
                        position but the user value of the motors concerned

<motors>
    <object role="role1" ... :list of motors to be moved to reach a predefined
    <object role="role2" ...  position. the "role" will be used to referenced
        ...                   the motors in the definitions of the predefined
<motors>                      positions

<deltas>                    : for each motor you define the windows used to
    <role1>val1</role1>       determine that a motor as reach a position
    <role2>val2</role2>
    ...
</deltas>

<positions>
    <position>
        <name>      : name of a predefined position. Must be unique in the file
        <role1>val1 : position of the motor "role1" for the predefined position
                     "name"
        <role2>val2 : position of the motor "role2" for the predefined position
                     "name"
        <resoy>8.69565217391e-07</resoy> : for all the position, independent
        <beamx>100</beamx>                 value with keyword can be added,
                                           saved, read ...
    </position>
    ...
</position>


METHOD
    name:           get_state
    input par.:     None
    output par.:    state
    description:    return an and on the state of all the  motor used in the
                    object

    name:           moveToPosition
    input par.:     name
    output par.:    None
    description:    move all motors to the predefined position "position"

    name:           get_value
    input par.:     None
    output par.:    position
    description:    return the name of the current predefined position.
                    return None if all motors are not in their psotion

    name:           setNewPositions
    input par.:     name, newPositions
    output par.:    None
    description:    For the position "name", change the motors positions set
                    in "newPositions", a dictionary with motor role as keys
                    and new motor position as values.
                    Save the new values in the xml file

    name:           getPositionKeyValue
    input par.:     name, key
    output par.:    value
    description:    return the value of the independent "key" field of
                    the predefined position "name"

    name:           setPositionKeyValue
    input par.:     name, key, value
    output par.:    None
    description:    Change in the object and in the xml file the value of the
                    independent field "key" in the predefined position "name"

    name:           getRoles
    input par.:     None
    output par.:    roles[]
    description:    return the list of motor's role used in the objects


SIGNAL
    name:           stateChanged
    parameter:      state
    description:    send the new state of the object when it changes

    name:           noPosition
    parameter:      None
    description:    sent when after a position change of any of the motor
                    the object is not in any of the predefined positions

    name:           positionReached
    parameter:      positionName
    description:    sent when after a position change of any of the motor
                    the object has reach a predefined position.
                    The parameter is the name of this position.

TEMPLATE
<object class="MultiplePositions">
    <username>VLM Zoom</username>
    <mode>absolute</mode>
    <motors>
        <object role="zoom" hwrid="/berru/zoom"></object>
    </motors>

    <deltas>
        <zoom>0.1</zoom>
    </deltas>

    <positions>
        <position>
            <name>1X</name>
            <zoom>0</zoom>
            <resox>-4.16666666667e-07</resox>
            <resoy>7.35294117647e-07</resoy>
            <beamx>537</beamx>
            <beamy>313</beamy>
        </position>
        <position>
            <name>6X</name>
            <zoom>1</zoom>
            <resox>-5.52486187845e-07</resox>
            <resoy>8.69565217391e-07</resoy>
            <beamx>100</beamx>
            <beamy>100</beamy>
        </position>
        <position>
            <name>12X</name>
            <zoom>2</zoom>
            <resox>0.0000004</resox>
            <resoy>0.0000004</resoy>
            <beamx>200</beamx>
            <beamy>200</beamy>
        </position>
    </positions>
</object>"""

try:
    from xml.etree import cElementTree  # python2.5
except ImportError:
    import cElementTree

import logging

from mxcubecore.BaseHardwareObjects import HardwareObject


class MultiplePositions(HardwareObject):
    def init(self):
        try:
            self.mode
        except AttributeError:
            self.mode = "absolute"

        motors = self["motors"]
        self.roles = motors.get_roles()

        self.deltas = {}
        try:
            # WARNING self.deltas is a LINK to the INTERNAL properties dictionary
            # modifying it modifies the GLOBAL properties, not just the local copy
            # Maybe do self["deltas"].get_properties().copy()?
            self.deltas = self["deltas"].get_properties()
        except Exception:
            logging.getLogger().error("No deltas.")

        self.positions = {}
        self.positionsIndex = []
        try:
            positions = self["positions"]
        except Exception:
            logging.getLogger().error("No positions.")
        else:
            for position in positions:
                name = position.get_property("name")
                if name is not None:
                    self.positionsIndex.append(name)
                    self.positions[name] = {}

                    motpos = position.get_properties()
                    motroles = list(motpos.keys())

                    for role in self.roles:
                        self.positions[name][role] = motpos[role]
                else:
                    logging.getLogger().error("No name for position.")

        self.motors = {}
        for mot in self["motors"]:
            self.motors[mot.get_motor_mnemonic()] = mot
            self.connect(mot, "moveDone", self.checkPosition)
            self.connect(mot, "valueChanged", self.checkPosition)
            self.connect(mot, "stateChanged", self.stateChanged)

    def get_state(self):
        if not self.is_ready():
            return ""

        state = "READY"
        for mot in self.motors.values():
            if mot.get_state() == mot.MOVING:
                state = "MOVING"
            elif mot.get_state() == mot.UNUSABLE:
                return "UNUSABLE"

        return state

    def stateChanged(self, state):
        self.emit("stateChanged", (self.get_state(),))
        self.checkPosition()

    def moveToPosition(self, name, wait=False):
        move_list = []
        for role in self.roles:
            device = self.get_deviceby_role(role)
            pos = self.positions[name][role]
            move_list.append((device, pos))

        for mot, pos in move_list:
            if mot is not None:
                mot.set_value(pos)

        if wait:
            [mot.waitEndOfMove() for mot, pos in move_list if mot is not None]
        """
        for mne,pos in self.positions[name].items():
        self.motors[mne].set_value(pos)
        """

    def get_value(self):
        if not self.is_ready():
            return None

        for posName, position in self.positions.items():
            findPosition = 0

            for role in self.roles:
                pos = position[role]
                mot = self.get_deviceby_role(role)

                if mot is not None:
                    motpos = mot.get_value()
                    try:
                        if (
                            motpos < pos + self.deltas[role]
                            and motpos > pos - self.deltas[role]
                        ):
                            findPosition += 1
                    except Exception:
                        continue

            if findPosition == len(self.roles):
                return posName

        return None

    def checkPosition(self, *args):
        if not self.is_ready():
            return None

        posName = self.get_value()

        if posName is None:
            self.emit("noPosition", ())
            return None
        else:
            self.emit("positionReached", (posName,))
            return posName

    def setNewPositions(self, name, newPositions):
        position = self.__getPositionObject(name)

        if position is None:
            self.checkPosition()
            return

        for role, pos in list(newPositions.items()):
            self.positions[name][role] = pos
            position.setProperty(role, pos)

        self.checkPosition()
        self.commit_changes()

    def getPositionKeyValue(self, name, key):
        position = self.__getPositionObject(name)

        if position is None:
            return None

        return position.get_property(key)

    def setPositionKeyValue(self, name, key, value):
        xml_tree = cElementTree.fromstring(self.xml_source())
        positions = xml_tree.find("positions")

        pos_list = positions.findall("position")
        for pos in pos_list:
            if pos.find("name").text == name:
                if pos.find(key) is not None:
                    position = self.__getPositionObject(name)
                    position.setProperty(key, str(value))
                    self.commit_changes()
                    return True
                else:
                    key_el = cElementTree.SubElement(pos, key)
                    key_el.text = value
                    print((cElementTree.tostring(xml_tree)))
                    self.rewrite_xml(cElementTree.tostring(xml_tree))
                    return True

        return False

    def __getPositionObject(self, name):
        for position in self["positions"]:
            if position.get_property("name") == name:
                return position

        return None

    def getRoles(self):
        return self.roles

    def addPosition(self, el_dict):
        xml_tree = cElementTree.fromstring(self.xml_source())
        positions = xml_tree.find("positions")

        pos = cElementTree.SubElement(positions, "position")

        for key, val in el_dict.items():
            sel = cElementTree.SubElement(pos, key)
            sel.text = val

        self.rewrite_xml(cElementTree.tostring(xml_tree))

    def remPosition(self, name):
        xml_tree = cElementTree.fromstring(self.xml_source())
        positions = xml_tree.find("positions")

        pos_list = positions.findall("position")
        for pos in pos_list:
            if pos.find("name").text == name:
                positions.remove(pos)

        self.rewrite_xml(cElementTree.tostring(xml_tree))

    def addField(self, name, key, val):
        pass

    def remField(self, name, key):
        pass


"""
        xml_tree = cElementTree.fromstring(self.xml_source())
        for elt in xml_tree.findall(".//position"):
           if elt.find("name").text=="12X":
             new_elt = cElementTree.Element("bidule")
             new_elt.text = "HELLO"
             elt.append(new_elt)
        self.rewrite_xml(cElementTree.tostring(xml_tree))
"""



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import time

import gevent
import requests

from mxcubecore import Poller
from mxcubecore.CommandContainer import CommandObject

__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"


class BlueskyHttpServerCommand(CommandObject):
    """Interface for communicating with the Bluesky Http Server API."""

    _default_timeout = 5
    _status_path = "api/status"
    _execute_path = "api/queue/item/execute"
    _console_uid_path = "api/console_output/uid"
    _console_path = "api/console_output_update"

    def __init__(self, name, url, timeout=5, **kwargs):
        CommandObject.__init__(self, name, **kwargs)
        self.username = ""
        self._default_timeout = timeout
        self._url = f"{url}/" if url[-1] != "/" else url
        self._headers = {"Authorization": f"ApiKey {os.environ['AUTH_KEY']}"}
        self.console_output_uid = self.get_console_uid()
        self.last_msg_uid = ""
        self.user_level_log = logging.getLogger("user_level_log")
        self.output_poller = Poller.poll(
            polled_call=self.update_console_output,
            polling_period=500,
            value_changed_callback=self.show_console_output,
            error_callback=self.poll_failed,
        )

    def poll_failed(self, exception, poller_id):
        self.user_level_log.error(exception)
        poller = Poller.get_poller(poller_id)
        if poller is not None:
            poller.restart(1000)

    def format_response(self, response):
        if response:
            response.raise_for_status()
            return json.loads(response.text)
        return response

    def status(self):
        response = requests.get(
            self._url + self._status_path,
            headers=self._headers,
            timeout=self._default_timeout,
        )
        return self.format_response(response)

    def monitor_manager_state(self, stop_state, timeout=86400):
        with gevent.Timeout(timeout, exception=TimeoutError):
            while self.status()["manager_state"] != stop_state:
                time.sleep(0.1)

    def execute_plan(self, plan_name, kwargs=None):
        if not kwargs:
            kwargs = {}
        return requests.post(
            self._url + self._execute_path,
            headers=self._headers,
            json={
                "user": self.username,
                "item": {"name": plan_name, "item_type": "plan", "kwargs": kwargs},
            },
            timeout=self._default_timeout,
        )

    def is_connected(self):
        http_server_status = self.status()
        re_environment_open = http_server_status["worker_environment_exists"]
        re_running = http_server_status["re_state"] is not None
        return re_environment_open and re_running

    def get_console_uid(self):
        response = requests.get(
            self._url + self._console_uid_path,
            headers=self._headers,
            timeout=self._default_timeout,
        )
        return self.format_response(response)["console_output_uid"]

    def show_console_output(self, value=None):
        if value is None:
            value = []
        for console_msg in value:
            new_console_line = console_msg["msg"]
            new_console_line = new_console_line.strip()
            if new_console_line != "":
                is_error = "[E " in new_console_line
                is_warning = "[W " in new_console_line
                is_debug = ("[I " in new_console_line) or ("[D " in new_console_line)
                if is_error:
                    self.user_level_log.error(new_console_line)
                elif is_warning:
                    self.user_level_log.warning(new_console_line)
                elif not is_debug:
                    self.user_level_log.info(new_console_line)

    def update_console_output(self):
        current_uid = self.get_console_uid()
        if self.console_output_uid != current_uid:
            response = requests.get(
                self._url + self._console_path,
                headers=self._headers,
                json={"last_msg_uid": self.last_msg_uid},
                timeout=self._default_timeout,
            )
            output = self.format_response(response)
            self.last_msg_uid = output["last_msg_uid"]
            self.console_output_uid = current_uid
            return output["console_output_msgs"]
        return []



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.model import queue_model_objects
from mxcubecore.queue_entry.base_queue_entry import BaseQueueEntry

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class XrayCentering2QueueEntry(BaseQueueEntry):
    """
    Entry for X-ray centring (2022 version)
    """

    def __init__(self, view=None, data_model=None):
        BaseQueueEntry.__init__(self, view, data_model)

    def execute(self):
        BaseQueueEntry.execute(self)
        HWR.beamline.xray_centring.execute()

    def pre_execute(self):
        """Pre-execute. Set to new motor position, if any"""
        BaseQueueEntry.pre_execute(self)
        logging.getLogger("user_level_log").info("Starting Xray centring, please wait.")

        data_model = self.get_data_model()

        motor_positions = dict(
            item
            for item in data_model.get_motor_positions().items()
            if item[1] is not None
        )
        pos_dict = {}
        for tag in ("kappa", "kappa_phi"):
            if tag in motor_positions:
                pos_dict[tag] = motor_positions.pop(tag)
        if pos_dict:
            # Some beamlines move centering motors while moving kappa, kappa_phi
            # Hence we need to move kappa, kappa_phi first.
            HWR.beamline.diffractometer.set_value_motors(pos_dict)
        if motor_positions:
            # Move the rest of the motors, if needed
            HWR.beamline.diffractometer.set_value_motors(motor_positions)

        HWR.beamline.xray_centring.pre_execute(self)

    def post_execute(self):
        """Post-execute. Store centring result in data model"""
        BaseQueueEntry.post_execute(self)

        # Create a centred position object of the current position
        # and put it in the data model for future access.
        pos_dict = HWR.beamline.diffractometer.get_positions()
        cpos = queue_model_objects.CentredPosition(pos_dict)
        self._data_model.set_centring_result(cpos)

        logging.getLogger("user_level_log").info("Finishing Xray centring")

        HWR.beamline.xray_centring.post_execute()

    def get_type_str(self):
        return "X-ray centring"



#  Project name: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""
Session hardware object.

Contains information regarding the current session and methods to
access and manipulate this information.
"""

import os
import time

from mxcubecore.HardwareObjects.Session import Session

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLSession(Session):
    """
    EMBLSession
    """

    def __init__(self, name):
        Session.__init__(self, name)
        self.start_time = time.strftime("%Y%m%d")

    def get_base_data_directory(self):
        """
        Returns the base data directory taking the 'contextual'
        information into account, such as if the current user
        is inhouse.

        :returns: The base data path.
        :rtype: str
        """
        user = os.getenv("USER").strip().lower()
        return os.path.join(
            self.base_directory,
            user,
            self.start_time,
        )

    def get_base_process_directory(self):
        """
        Returns base process directory
        :return: str
        """
        user = os.getenv("USER").lower()

        process_directory = os.path.join(
            self.base_process_directory, user, self.start_time, "PROCESSED_DATA"
        )
        return process_directory.replace(" ", "")
